// Standard includes
#include <unistd.h>
#include <errno.h>
#include <string.h>

// Vendor includes

// Project includes
#include "Logger.hpp"

// Using declarations


LoggerServer::LoggerServer()
    : m_IsRunning( false ),
      m_IsTerminationRequested( false )
{
}

LoggerServer::~LoggerServer()
{
    this->ScheduleClose();
}

bool LoggerServer::Open()
{
    if ( !m_IsRunning )
    {
        if ( 0 == ::pipe( &( m_Pipe[ 0 ] ) ) )
        {
            m_IsTerminationRequested = false;
            if ( this->ThreadActivate() )
            {
                m_IsRunning = true;
            }
            else
            {
                ::close( m_Pipe[ PIPE_CHANNEL__READ ] );
                ::close( m_Pipe[ PIPE_CHANNEL__WRITE ] );
            }
        }
    }

    return m_IsRunning;
}

void LoggerServer::ScheduleClose()
{
    m_IsTerminationRequested = true;

    // Unblock the receive
    ::write( m_Pipe[ PIPE_CHANNEL__WRITE ], &( m_MessageBuffer[ 0 ] ), sizeof ( m_MessageBuffer ) );
}

bool LoggerServer::IsRunning() const
{
    return m_IsRunning;
}

void LoggerServer::Run( void* pUserData )
{
    MessageHeader* pMessageHeader = reinterpret_cast < MessageHeader* > ( &( m_MessageBuffer[ 0 ] ) );

    while ( true )
    {
        // Read message size
        if ( !this->ReadMessagePart( 0, sizeof ( MessageHeader ) ) ||
             !this->ReadMessagePart( sizeof ( MessageHeader ), pMessageHeader->MessageBodyLength ) )
        {
            break;
        }

        this->Output( pMessageHeader );
    }

    this->Flush();

    m_IsRunning = false;
}

bool LoggerServer::ReadMessagePart( unsigned int startBufferOffset, unsigned int bytesToRead )
{
    unsigned int currentOffset = 0;

    while ( bytesToRead != 0 )
    {
        int bytesRead = ::read( m_Pipe[ PIPE_CHANNEL__READ ], &( m_MessageBuffer[ startBufferOffset + currentOffset ] ), bytesToRead );
        if ( ( 0 == bytesRead ) || ( ( bytesRead < 0 ) && ( EINTR != errno ) ) || m_IsTerminationRequested )
        {
            // The pipe was closed or read fails not because of signal
            ::close( m_Pipe[ PIPE_CHANNEL__READ ] );
            ::close( m_Pipe[ PIPE_CHANNEL__WRITE ] );
            break;
        }

        if ( bytesRead > 0 )
        {
            currentOffset += bytesRead;
            bytesToRead -= bytesRead;
        }
    }

    return ( 0 == bytesToRead );
}

void LoggerServer::WriteMessage( MessageHeader* pMessageHeader )
{
    // Internal function so parameters are not checked
    ::write( m_Pipe[ PIPE_CHANNEL__WRITE ], pMessageHeader, sizeof ( MessageHeader ) + pMessageHeader->MessageBodyLength );
}

Manip::Manip()
    : m_Base( 10 ),
      m_Fill( ' ' ),
      m_Width( 0 )
{
}

Manip::Manip( unsigned char base, char fill, unsigned char width )
    : m_Fill( fill ),
      m_Width( width )
{
    switch ( base )
    {
    case 2:
    case 8:
    case 10:
    case 16:
        m_Base = base;
        break;

    default:
        m_Base = 10;
        break;
    }
}

Manip::Manip( const Manip& rOther )
    : m_Base( rOther.m_Base ),
      m_Fill( rOther.m_Fill ),
      m_Width( rOther.m_Width )
{
}

Manip::~Manip()
{
}

Manip& Manip::operator=( const Manip& rOther )
{
      m_Base = rOther.m_Base;
      m_Fill = rOther.m_Fill;
      m_Width = rOther.m_Width;

      return *this;
}

unsigned char Manip::GetBase() const
{
    return m_Base;
}

char Manip::GetFill() const
{
    return m_Fill;
}

unsigned char Manip::GetWidth() const
{
    return m_Width;
}

LoggerClient::LoggerClient( unsigned char clientId )
    : m_pLoggerServer( 0 ),
      m_LoggerLevel( LOGGER_LEVEL__ERROR ),
      m_CurrentSectionLoggerLevel( LOGGER_LEVEL__ERROR ),
      m_CurrentMessageBodyOffset( 0 ),
      m_MessageBufferCapacity( sizeof ( m_MessageBuffer ) - sizeof ( LoggerServer::MessageHeader ) - 1 ) // One extra symbol for terminating zero
{
    reinterpret_cast < LoggerServer::MessageHeader* > ( &( m_MessageBuffer[ 0 ] ) )->ClientId = clientId;
}

LoggerClient::~LoggerClient()
{
}

void LoggerClient::AttachLoggerServer( LoggerServer* pLoggerServer )
{
    // Could be re-attached as well
    m_pLoggerServer = pLoggerServer;
}

void LoggerClient::SetLevel( LoggerLevel loggerLevel )
{
    switch ( loggerLevel )
    {
    case LOGGER_LEVEL__ERROR:
    case LOGGER_LEVEL__WARNING:
    case LOGGER_LEVEL__INFO:
    case LOGGER_LEVEL__DEBUG:
        m_LoggerLevel = loggerLevel;
        break;

    default:
        m_LoggerLevel = LOGGER_LEVEL__ERROR;
        break;
    }
}

bool LoggerClient::OpenSection( LoggerLevel sectionLoggerLevel )
{
    m_CurrentSectionLoggerLevel = sectionLoggerLevel;

    return ( sectionLoggerLevel <= m_LoggerLevel );
}

LoggerClient& LoggerClient::operator>>( Manip& rManip )
{
    rManip = m_Manip;

    return *this;
}

LoggerClient& LoggerClient::operator<<( Manip manip )
{
    m_Manip = manip;

    return *this;
}

LoggerClient& LoggerClient::operator<<( LoggerClient &( *parameterlessManip )( LoggerClient& rLoggerClient ) )
{
    return parameterlessManip( *this );
}

LoggerClient& LoggerClient::operator<<( char* pString )
{
    if ( 0 != pString )
    {
        unsigned int stringLength = ::strlen( pString );

        if ( m_MessageBufferCapacity < stringLength )
        {
            stringLength = m_MessageBufferCapacity;
        }
        if ( stringLength > 0 )
        {
            ::memcpy( &( m_MessageBuffer[ sizeof ( LoggerServer::MessageHeader ) + m_CurrentMessageBodyOffset ] ), pString, stringLength );
            m_CurrentMessageBodyOffset += stringLength;
            m_MessageBufferCapacity -= stringLength;
        }
    }

    return *this;
}

LoggerClient& LoggerClient::operator<<( char symbol )
{
    if ( m_MessageBufferCapacity >= 1 )
    {
        m_MessageBuffer[ sizeof ( LoggerServer::MessageHeader ) + m_CurrentMessageBodyOffset++ ] = symbol;
        m_MessageBufferCapacity--;
    }

    return *this;
}

LoggerClient& LoggerClient::operator<<( unsigned int value )
{
    char tempBuffer[ 8 * sizeof ( value ) ];
    unsigned char tempBuferIndex = sizeof ( tempBuffer );

    do
    {
        tempBuffer[ --tempBuferIndex ] = "0123456789ABCDEF"[ value % m_Manip.GetBase() ];
        value = value / m_Manip.GetBase();
    }
    while ( value > 0 );

    // Get real value and fill lengths
    int valueLength = sizeof ( tempBuffer ) - tempBuferIndex;
    int fillLength = ( m_Manip.GetWidth() > valueLength ) ? ( m_Manip.GetWidth() - valueLength ) : 0;

    // Adjust real value and fill lengths
    if ( static_cast < int > ( m_MessageBufferCapacity ) >= fillLength )
    {
        if ( static_cast < int > ( m_MessageBufferCapacity ) < ( fillLength + valueLength ) )
        {
            valueLength = m_MessageBufferCapacity - fillLength;
        }
    }
    else
    {
        valueLength = 0;
        fillLength = m_MessageBufferCapacity;
    }

    // Write to the output buffer
    char* pCurrentMessageBody = &( m_MessageBuffer[ sizeof ( LoggerServer::MessageHeader ) + m_CurrentMessageBodyOffset ] );
    m_CurrentMessageBodyOffset += ( fillLength + valueLength );
    m_MessageBufferCapacity -= ( fillLength + valueLength );
    while ( fillLength-- > 0 )
    {
        *pCurrentMessageBody = m_Manip.GetFill();
        pCurrentMessageBody++;
    }
    while ( valueLength-- > 0 )
    {
        *pCurrentMessageBody = tempBuffer[ tempBuferIndex++ ];
        pCurrentMessageBody++;
    }

    return *this;
}

LoggerClient& LoggerClient::operator<<( int value )
{
    if ( m_MessageBufferCapacity >= 1 )
    {
        // Output sign for base 10 only
        if ( 10 == m_Manip.GetBase() )
        {
            if ( value < 0 )
            {
                this->operator <<( '-' );
                value = -value;
            }
            else
            {
                this->operator <<( '+' );
            }
        }

        this->operator<<( UINT ( value ) );
    }

    return *this;
}

void LoggerClient::Flush()
{
    if ( m_CurrentMessageBodyOffset > 0 )
    {
        LoggerServer* pLoggerServer = m_pLoggerServer;
        if ( 0 != pLoggerServer )
        {
            // Add terminating zero
            m_MessageBuffer[ sizeof ( LoggerServer::MessageHeader ) + m_CurrentMessageBodyOffset++ ] = 0;
            reinterpret_cast < LoggerServer::MessageHeader* > ( &( m_MessageBuffer[ 0 ] ) )->Level = m_CurrentSectionLoggerLevel;
            reinterpret_cast < LoggerServer::MessageHeader* > ( &( m_MessageBuffer[ 0 ] ) )->MessageBodyLength = m_CurrentMessageBodyOffset;
            pLoggerServer->WriteMessage( reinterpret_cast < LoggerServer::MessageHeader* > ( &( m_MessageBuffer[ 0 ] ) ) );
        }

        m_CurrentMessageBodyOffset = 0;
        m_MessageBufferCapacity = sizeof ( m_MessageBuffer ) - sizeof ( LoggerServer::MessageHeader ) - 1; // One extra symbol for terminating zero
    }
}

LoggerClient& Endl( LoggerClient& rLoggerClient )
{
    return ( rLoggerClient << '\n' );
}
