#ifndef LOGGER__HPP
#define LOGGER__HPP

// Standard includes

// Vendor includes

// Project includes
#include "Thread.hpp"


// Type conversion for LoggerClient
#define STRING( _pString ) ( ( char* )( _pString ) )
#define UINT( _value ) ( static_cast < unsigned int > ( _value ) )
#define INT( _value ) ( static_cast < int > ( _value ) )


class LoggerServer : public Thread
{
public:
    friend class LoggerClient; // Do not want to expose internal interfaces

    LoggerServer();
    virtual ~LoggerServer();

    bool Open();
    void ScheduleClose();
    bool IsRunning() const;

protected:
    struct MessageHeader
    {
        unsigned char ClientId;
        unsigned char Level;
        unsigned short MessageBodyLength;
    } __attribute__ ( ( __packed__ ) );

    virtual void Output( MessageHeader* pMessage ) = 0;
    virtual void Flush() = 0;

private:
    enum
    {
        // Control pipe identifiers
        PIPE_CHANNEL__READ = 0,
        PIPE_CHANNEL__WRITE,
        PIPE_CHANNEL__MAX
    };

    enum
    {
        // For a better user experience this value should never exceed
        // the maximum pipe atomic write size ( PIPE_BUF in limits.h )
        LOG_MESSAGE_BUFFER_SIZE_MAX = 512
    };

    // Copy constructor is invalid and not defined, declared here to
    // prevent the compiler generating the default definition
    LoggerServer( const LoggerServer& rOther );

    // Copy assignment operator is invalid and not defined, declared
    // here to prevent compiler generating the default definition
    LoggerServer& operator=( const LoggerServer& rOther );

    virtual void Run( void* pUserData );
    bool ReadMessagePart( unsigned int startBufferOffset, unsigned int bytesToRead );
    void WriteMessage( MessageHeader* pMessageHeader );

    volatile bool m_IsRunning;
    volatile bool m_IsTerminationRequested;
    int m_Pipe[ PIPE_CHANNEL__MAX ];
    char m_MessageBuffer[ LOG_MESSAGE_BUFFER_SIZE_MAX ];
};

class Manip
{
public:
    Manip();
    Manip( unsigned char base, char fill, unsigned char width );
    Manip( const Manip& rOther );
    ~Manip(); // Do not make this destructor virtual

    Manip& operator=( const Manip& rOther );
    unsigned char GetBase() const;
    char GetFill() const;
    unsigned char GetWidth() const;

protected:

private:
    unsigned char m_Base;
    char m_Fill;
    unsigned char m_Width;
} __attribute__ ( ( __packed__ ) );

class LoggerClient
{
public:
    enum LoggerLevel
    {
        LOGGER_LEVEL__ERROR = 0,
        LOGGER_LEVEL__WARNING,
        LOGGER_LEVEL__INFO,
        LOGGER_LEVEL__DEBUG
    };

    LoggerClient( unsigned char clientId );
    virtual ~LoggerClient();

    void AttachLoggerServer( LoggerServer* pLoggerServer );
    void SetLevel( LoggerLevel loggerLevel );
    bool OpenSection( LoggerLevel sectionLoggerLevel );
    LoggerClient& operator>>( Manip& rManip );
    LoggerClient& operator<<( Manip manip );
    LoggerClient& operator<<( LoggerClient &( *parameterlessManip )( LoggerClient& rLoggerClient ) );
    LoggerClient& operator<<( char* pString );
    LoggerClient& operator<<( char symbol );
    LoggerClient& operator<<( unsigned int value );
    LoggerClient& operator<<( int value );
    void Flush();

protected:

private:
    // Copy constructor is invalid and not defined, declared here to
    // prevent the compiler generating the default definition
    LoggerClient( const LoggerClient& rOther );

    // Copy assignment operator is invalid and not defined, declared
    // here to prevent compiler generating the default definition
    LoggerClient& operator=( const LoggerClient& rOther );

    Manip m_Manip;
    LoggerServer* m_pLoggerServer;
    LoggerLevel m_LoggerLevel;
    LoggerLevel m_CurrentSectionLoggerLevel;
    char m_MessageBuffer[ LoggerServer::LOG_MESSAGE_BUFFER_SIZE_MAX ];
    unsigned int m_CurrentMessageBodyOffset;
    unsigned int m_MessageBufferCapacity;
};

LoggerClient& Endl( LoggerClient& rLoggerClient );


#endif // LOGGER__HPP
