// Standard includes
#include <string.h>
#include <stdio.h>

// Vendor includes

// Project includes
#include "ReaderBase.hpp"
#include "DallasCrc8.hpp"

// Using declarations


ReaderBase::ReaderBase( LoggerClient& rLoggerClient, unsigned char* pCommandPacket )
    : m_rLoggerClient( rLoggerClient ),
      m_pNode( 0 ),
      m_SerialNumber( 0 ),
      m_pCommandPacket( pCommandPacket ),
      m_ReadTimeoutInMilliseconds( DEFAULT_READ_TIMEOUT_MILLISECONDS )
{
    PacketHeader* pPacketHeader = reinterpret_cast < PacketHeader* > ( m_pCommandPacket );
    pPacketHeader->StartOfPacket = PACKET__START;
    pPacketHeader->TargetStationId = PACKET__TARGET_STATION_ID;
    pPacketHeader->SourceStationId = PACKET__SOURCE_STATION_ID;
    pPacketHeader->ProtocolOptionCode = PACKET__PROTOCOL_OPTION_CODE;

    m_PacketTail.EndOfPacket = PACKET__END;
}

ReaderBase::~ReaderBase()
{
    this->Close();
}

const char* ReaderBase::GetNode() const
{
    return m_pNode;
}

unsigned long ReaderBase::GetSerialNumber() const
{
    return m_SerialNumber;
}

void ReaderBase::LogIdentity( LoggerClient& rLoggerClient )
{
    rLoggerClient << STRING( m_pNode ) << STRING( ", S/N " ) << Manip( 10, ' ', 0 ) << UINT( m_SerialNumber );
}

bool ReaderBase::Open( const char* pNode, unsigned long serialNumber, unsigned int speed, char parityCode, unsigned char stopBits )
{
    bool isSuccess = false;

    if ( 0 == m_pNode )
    {
        if ( 0 != pNode )
        {
            unsigned int nodeLength = ::strlen( pNode );
            m_pNode = new char [ nodeLength + 1 ];
            if ( 0 != m_pNode )
            {
                ::memcpy( m_pNode, pNode, nodeLength + 1 );

                if ( m_UsbSerial.Open( m_pNode, speed, parityCode, stopBits ) )
                {
                    m_SerialNumber = serialNumber;

                    m_PortConfiguration.Speed = speed;
                    m_PortConfiguration.ParityCode = parityCode;
                    m_PortConfiguration.StopBits = stopBits;

                    m_ReadTimeoutInMilliseconds = -1;

                    if ( m_rLoggerClient.OpenSection( LoggerClient::LOGGER_LEVEL__INFO ) )
                    {
                        m_rLoggerClient << STRING( "Reader [ " );
                        this->LogIdentity( m_rLoggerClient );
                        m_rLoggerClient << STRING( " ] opened: " ) << UINT( speed ) << '-' << parityCode << '-' << UINT( stopBits ) << Endl;
                        m_rLoggerClient.Flush();
                    }

                    isSuccess = true;
                }
                else if ( m_rLoggerClient.OpenSection( LoggerClient::LOGGER_LEVEL__ERROR ) )
                {
                    delete [] m_pNode;
                    m_pNode = 0;

                    m_rLoggerClient << STRING( "Can't open reader [ " );
                    this->LogIdentity( m_rLoggerClient );
                    m_rLoggerClient << STRING( " ]: " ) << STRING( m_UsbSerial.GetLastErrorString() ) << Endl;
                    m_rLoggerClient.Flush();
                }
            }
        }
    }
    else if ( m_rLoggerClient.OpenSection( LoggerClient::LOGGER_LEVEL__WARNING ) )
    {
        m_rLoggerClient << STRING( "Reader [ " );
        this->LogIdentity( m_rLoggerClient );
        m_rLoggerClient << STRING( " ] already opened: " ) << UINT( m_PortConfiguration.Speed ) << '-' << m_PortConfiguration.ParityCode << '-'
                        << UINT( m_PortConfiguration.StopBits ) << Endl;
        m_rLoggerClient.Flush();
    }

    return isSuccess;
}

bool ReaderBase::Reopen()
{
    bool isSuccess = false;

    m_UsbSerial.Close();

    if ( ( 0 != m_pNode ) && m_UsbSerial.Open( m_pNode, m_PortConfiguration.Speed, m_PortConfiguration.ParityCode, m_PortConfiguration.StopBits ) )
    {
        if ( m_rLoggerClient.OpenSection( LoggerClient::LOGGER_LEVEL__INFO ) )
        {
            m_rLoggerClient << STRING( "Reader [ " );
            this->LogIdentity( m_rLoggerClient );
            m_rLoggerClient << STRING( " ] re-opened" ) << Endl;
            m_rLoggerClient.Flush();
        }

        isSuccess = true;
    }
    else if ( m_rLoggerClient.OpenSection( LoggerClient::LOGGER_LEVEL__ERROR ) )
    {
        m_rLoggerClient << STRING( "Can't re-open reader [ " );
        this->LogIdentity( m_rLoggerClient );
        m_rLoggerClient << STRING( " ]: " ) << STRING( m_UsbSerial.GetLastErrorString() ) << Endl;
        m_rLoggerClient.Flush();
    }

    return isSuccess;
}

void ReaderBase::Close()
{
    if ( 0 != m_pNode )
    {
        m_UsbSerial.Close();

        if ( m_rLoggerClient.OpenSection( LoggerClient::LOGGER_LEVEL__INFO ) )
        {
            m_rLoggerClient << STRING( "Reader [ " );
            this->LogIdentity( m_rLoggerClient );
            m_rLoggerClient << STRING( " ] closed" ) << Endl;
            m_rLoggerClient.Flush();
        }

        delete [] m_pNode;
        m_pNode = 0;
        m_SerialNumber = 0;
    }
}

void ReaderBase::SetReadTimeout( int readTimeoutInMilliseconds )
{
    m_ReadTimeoutInMilliseconds = readTimeoutInMilliseconds;
}

bool ReaderBase::ProcessCommand( unsigned char* pCommandBuffer, unsigned char* pResponseBuffer, unsigned int responseBufferSize )
{
    bool isSuccess = false;

    if ( 0 != m_pNode )
    {
        CommandHeader* pCommandHeader = reinterpret_cast < CommandHeader* > ( pCommandBuffer );
        if ( ( 0 != pCommandBuffer ) && ( 0 != pResponseBuffer ) && ( responseBufferSize >= sizeof ( ResponseHeader ) ) )
        {
            // Copy payload, calculate CRC and copy packet tail
            unsigned int commandPacketLength = sizeof ( PacketHeader );
            ::memcpy( m_pCommandPacket + commandPacketLength, pCommandBuffer, sizeof ( CommandHeader ) + pCommandHeader->Length );
            commandPacketLength += sizeof ( CommandHeader ) + pCommandHeader->Length;
            DallasCrc8 dallasCrc8;
            m_PacketTail.Crc8 = dallasCrc8.AddBuffer( m_pCommandPacket, commandPacketLength );
            *( reinterpret_cast < PacketTail* > ( m_pCommandPacket + commandPacketLength ) ) = m_PacketTail;
            commandPacketLength += sizeof ( PacketTail );

            // Write the command to the reader
            if ( m_UsbSerial.Write( m_pCommandPacket, commandPacketLength ) )
            {
                if ( m_rLoggerClient.OpenSection( LoggerClient::LOGGER_LEVEL__DEBUG ) )
                {
                    m_rLoggerClient << STRING( "Command sent to reader [ " );
                    this->LogIdentity( m_rLoggerClient );
                    m_rLoggerClient << STRING( " ]: " ) << Manip( 16, '0', 2 );
                    for ( unsigned int i = 0; i < commandPacketLength; i++ )
                    {
                        m_rLoggerClient << STRING( " 0x" ) << UINT( *( m_pCommandPacket + i ) );
                    }
                    m_rLoggerClient << Endl;
                    m_rLoggerClient.Flush();
                }

                // Wait for reader response
                PacketHeader packetHeader;
                if ( m_UsbSerial.Read( reinterpret_cast < unsigned char* > ( &packetHeader ), sizeof ( packetHeader ), m_ReadTimeoutInMilliseconds ) )
                {
                    // Compare with the header template
                    if ( 0 == ::memcmp( &packetHeader, m_pCommandPacket, sizeof ( packetHeader ) ) )
                    {
                        if ( m_UsbSerial.Read( pResponseBuffer, sizeof ( ResponseHeader ), m_ReadTimeoutInMilliseconds ) )
                        {
                            unsigned int ResponseBodyLength = ( reinterpret_cast < ResponseHeader* > ( pResponseBuffer ) )->Length;
                            if ( responseBufferSize >= ( sizeof ( ResponseHeader ) + ResponseBodyLength ) )
                            {
                                isSuccess = true;

                                // Read response packet body if exists
                                if ( ResponseBodyLength > 0 )
                                {
                                    isSuccess = false;

                                    if ( m_UsbSerial.Read( pResponseBuffer + sizeof ( ResponseHeader ), ResponseBodyLength, m_ReadTimeoutInMilliseconds ) )
                                    {
                                        isSuccess = true;
                                    }
                                    else if ( m_rLoggerClient.OpenSection( LoggerClient::LOGGER_LEVEL__ERROR ) )
                                    {
                                        m_rLoggerClient << STRING( "Reader [ " );
                                        this->LogIdentity( m_rLoggerClient );
                                        m_rLoggerClient << STRING( " ] receive packet body error: " ) << STRING( m_UsbSerial.GetLastErrorString() ) << Endl;
                                        m_rLoggerClient.Flush();
                                    }
                                }

                                // Read response packet tail
                                PacketTail packetTail;
                                if ( isSuccess )
                                {
                                    isSuccess = false;

                                    if ( m_UsbSerial.Read( reinterpret_cast < unsigned char* > ( &packetTail ), sizeof ( packetTail ),
                                                           m_ReadTimeoutInMilliseconds ) )
                                    {
                                        dallasCrc8.Restart();
                                        dallasCrc8.AddBuffer( reinterpret_cast < unsigned char* > ( &packetHeader ), sizeof ( packetHeader ) );
                                        m_PacketTail.Crc8 = dallasCrc8.AddBuffer( pResponseBuffer, sizeof ( ResponseHeader ) + ResponseBodyLength );
                                        if ( 0 == ::memcmp( &packetTail,  &m_PacketTail, sizeof ( packetTail ) ) )
                                        {
                                            isSuccess = true;
                                        }
                                        else if ( m_rLoggerClient.OpenSection( LoggerClient::LOGGER_LEVEL__ERROR ) )
                                        {
                                            m_rLoggerClient << STRING( "Reader [ " );
                                            this->LogIdentity( m_rLoggerClient );

                                            if ( packetTail.Crc8 != m_PacketTail.Crc8 )
                                            {
                                                m_rLoggerClient << STRING( " ] CRC error: expected 0x" ) << Manip( 16, '0', 2 ) << UINT( m_PacketTail.Crc8 )
                                                                << STRING( ", received 0x" ) << UINT( packetTail.Crc8 );
                                            }
                                            else
                                            {
                                                m_rLoggerClient << STRING( " ] packet tail error: " ) << Manip( 16, '0', 2 );
                                                for ( unsigned int i = 0; i < sizeof ( PacketHeader ); i++ )
                                                {
                                                    m_rLoggerClient << STRING( " 0x" ) << UINT( pResponseBuffer[ i ] );
                                                }
                                            }
                                            m_rLoggerClient << Endl;
                                            m_rLoggerClient.Flush();
                                        }
                                    }
                                    else if ( m_rLoggerClient.OpenSection( LoggerClient::LOGGER_LEVEL__ERROR ) )
                                    {
                                        m_rLoggerClient << STRING( "Reader [ " );
                                        this->LogIdentity( m_rLoggerClient );
                                        m_rLoggerClient << STRING( " ] packet tail read error: " ) << STRING( m_UsbSerial.GetLastErrorString() ) << Endl;
                                        m_rLoggerClient.Flush();
                                    }
                                }

                                if ( isSuccess && m_rLoggerClient.OpenSection( LoggerClient::LOGGER_LEVEL__DEBUG ) )
                                {
                                    m_rLoggerClient << STRING( "Response received from reader [ " );
                                    this->LogIdentity( m_rLoggerClient );
                                    m_rLoggerClient << STRING( " ]: " ) << Manip( 16, '0', 2 );
                                    for ( unsigned int i = 0; i < sizeof ( PacketHeader ); i++ )
                                    {
                                        m_rLoggerClient << STRING( " 0x" ) << UINT( *( reinterpret_cast < unsigned char* > ( &packetHeader ) + i ) );
                                    }
                                    for ( unsigned int i = 0; i < sizeof ( ResponseHeader ) + ResponseBodyLength; i++ )
                                    {
                                        m_rLoggerClient << STRING( " 0x" ) << UINT( *( pResponseBuffer + i ) );
                                    }
                                    for ( unsigned int i = 0; i < sizeof ( PacketTail ); i++ )
                                    {
                                        m_rLoggerClient << STRING( " 0x" ) << UINT( *( reinterpret_cast < unsigned char* > ( &packetTail ) + i ) );
                                    }
                                    m_rLoggerClient << Endl;
                                    m_rLoggerClient.Flush();
                                }
                            }
                            else if ( m_rLoggerClient.OpenSection( LoggerClient::LOGGER_LEVEL__ERROR ) )
                            {
                                m_rLoggerClient << STRING( "Reader [ " );
                                this->LogIdentity( m_rLoggerClient );
                                m_rLoggerClient << STRING( " ] receive error: provided buffer is too small" ) << Endl;
                                m_rLoggerClient.Flush();
                            }
                        }
                        else if ( m_rLoggerClient.OpenSection( LoggerClient::LOGGER_LEVEL__ERROR ) )
                        {
                            m_rLoggerClient << STRING( "Reader [ " );
                            this->LogIdentity( m_rLoggerClient );
                            m_rLoggerClient << STRING( " ] receive response header error: " ) << STRING( m_UsbSerial.GetLastErrorString() ) << Endl;
                            m_rLoggerClient.Flush();
                        }
                    }
                    else if ( m_rLoggerClient.OpenSection( LoggerClient::LOGGER_LEVEL__ERROR ) )
                    {
                        m_rLoggerClient << STRING( "Reader [ " );
                        this->LogIdentity( m_rLoggerClient );
                        m_rLoggerClient << STRING( " ] packet header error: " ) << Manip( 16, '0', 2 );
                        for ( unsigned int i = 0; i < sizeof ( PacketHeader ); i++ )
                        {
                            m_rLoggerClient << STRING( " 0x" ) << UINT( pResponseBuffer[ i ] );
                        }
                        m_rLoggerClient << Endl;
                        m_rLoggerClient.Flush();
                    }
                }
                else if ( m_rLoggerClient.OpenSection( LoggerClient::LOGGER_LEVEL__ERROR ) )
                {
                    m_rLoggerClient << STRING( "Reader [ " );
                    this->LogIdentity( m_rLoggerClient );
                    m_rLoggerClient << STRING( " ] receive packet header error: " ) << STRING( m_UsbSerial.GetLastErrorString() ) << Endl;
                    m_rLoggerClient.Flush();
                }
            }
            else if ( m_rLoggerClient.OpenSection( LoggerClient::LOGGER_LEVEL__WARNING ) )
            {
                m_rLoggerClient << STRING( "Reader [ " );
                this->LogIdentity( m_rLoggerClient );
                m_rLoggerClient << STRING( " ] transmit error: " ) << STRING( m_UsbSerial.GetLastErrorString() ) << Endl;
                m_rLoggerClient.Flush();
            }
        }
        else if ( m_rLoggerClient.OpenSection( LoggerClient::LOGGER_LEVEL__WARNING ) )
        {
            m_rLoggerClient << STRING( "Reader [ " );
            this->LogIdentity( m_rLoggerClient );
            m_rLoggerClient << STRING( " ] transmit error: parameters are not valid" ) << Endl;
            m_rLoggerClient.Flush();
        }
    }

    return isSuccess;
}
