// Standard includes
#include <unistd.h>
#include <string.h>

// Vendor includes

// Project includes
#include "LoggerServerImpl.hpp"
#include "ReaderFabric.hpp"
#include "ReaderModelPrp08.hpp"
#include "ReaderModelPrEh08.hpp"

// Using declarations


ReaderFabric::ReaderFabric()
    : m_LoggerClient( LoggerServerImpl::LOGGER_CLIENT_ID__READER_FABRIC )
{
}

ReaderFabric::~ReaderFabric()
{
    this->Terminate();
}

void ReaderFabric::SetLogger( LoggerServer* pLoggerServer, LoggerClient::LoggerLevel loggerLevel, LoggerClient::LoggerLevel plugAndPlayLoggerLevel )
{
    this->ReaderPlugAndPlay::SetLogger( pLoggerServer, plugAndPlayLoggerLevel );

    if ( 0 != pLoggerServer )
    {
        m_LoggerClient.AttachLoggerServer( pLoggerServer );
    }
    m_LoggerClient.SetLevel( loggerLevel );
}

void ReaderFabric::SetReaderLogger( LoggerServer* pLoggerServer, LoggerClient::LoggerLevel loggerLevel )
{
    m_ReaderLogger.pServer = pLoggerServer;
    m_ReaderLogger.Level = loggerLevel;
}

bool ReaderFabric::Initialize()
{
    bool isSuccess = false;

    if ( !this->ReaderPlugAndPlay::IsRunning() )
    {
        if ( this->ReaderPlugAndPlay::ThreadStart( 0 ) )
        {
            if ( this->ReaderPlugAndPlay::Open() )
            {
                ::memset( &( m_InterfaceDescriptor[ 0 ] ), 0, sizeof ( m_InterfaceDescriptor ) );
                ::pthread_mutex_init( &m_InterfaceDescriptorLock, 0 );

                isSuccess = true;;

                if ( m_LoggerClient.OpenSection( LoggerClient::LOGGER_LEVEL__DEBUG ) )
                {
                    m_LoggerClient << STRING( "Reader fabric initialized" ) << Endl;
                    m_LoggerClient.Flush();
                }

                if ( !isSuccess )
                {
                    this->Terminate();
                }
            }
        }
        else if ( m_LoggerClient.OpenSection( LoggerClient::LOGGER_LEVEL__ERROR ) )
        {
            m_LoggerClient << STRING( "Can't start device manager worker thread" ) << Endl;
            m_LoggerClient.Flush();
        }
    }
    else if ( m_LoggerClient.OpenSection( LoggerClient::LOGGER_LEVEL__WARNING ) )
    {
        m_LoggerClient << STRING( "Reader fabric already initialized" ) << Endl;
        m_LoggerClient.Flush();
    }

    return isSuccess;
}

void ReaderFabric::Terminate()
{
    if ( this->ReaderPlugAndPlay::IsRunning() )
    {
        this->ReaderPlugAndPlay::ScheduleClose();
        while ( this->ReaderPlugAndPlay::IsRunning() )
        {
            ::usleep( 100000 );
        }

        this->ReaderPlugAndPlay::ThreadStop();

        for ( unsigned int i = 0; i < sizeof ( m_InterfaceDescriptor ) / sizeof ( m_InterfaceDescriptor[ 0 ] ); i++ )
        {
            m_InterfaceDescriptor[ i ].SerialNumber = INVALID_SERIAL_NUMBER;
        }

        ::pthread_mutex_lock( &m_InterfaceDescriptorLock );

        for ( unsigned int i = 0; i < sizeof ( m_InterfaceDescriptor ) / sizeof ( m_InterfaceDescriptor[ 0 ] ); i++ )
        {
            this->RemoveDevice( &( m_InterfaceDescriptor[ i ] ) );
        }

        ::pthread_mutex_unlock( &m_InterfaceDescriptorLock );

        ::pthread_mutex_destroy( &m_InterfaceDescriptorLock );
    }
}

int ReaderFabric::LockInterface( unsigned long serialNumber, unsigned short& rProductId, ReaderBase*& rpReaderBase )
{
    int interfaceId;

    ::pthread_mutex_lock( &m_InterfaceDescriptorLock );

    for ( interfaceId = 0; interfaceId < static_cast < int > ( sizeof ( m_InterfaceDescriptor ) / sizeof ( m_InterfaceDescriptor[ 0 ] ) ); interfaceId++ )
    {
        InterfaceDescriptor* pCurrentInterfaceDescriptor = &( m_InterfaceDescriptor[ interfaceId ] );

        if ( pCurrentInterfaceDescriptor->SerialNumber == serialNumber )
        {
            rProductId = pCurrentInterfaceDescriptor->ProductId;
            rpReaderBase = pCurrentInterfaceDescriptor->pReader;

            pCurrentInterfaceDescriptor->IsLocked = true;

            if ( m_LoggerClient.OpenSection( LoggerClient::LOGGER_LEVEL__DEBUG ) )
            {
                m_LoggerClient << STRING( "Reader [ " );
                rpReaderBase->LogIdentity( m_LoggerClient );
                m_LoggerClient << STRING( " ] locked" ) << Endl;
                m_LoggerClient.Flush();
            }

            break;
        }
    }

    ::pthread_mutex_unlock( &m_InterfaceDescriptorLock );

    if ( ( sizeof ( m_InterfaceDescriptor ) / sizeof ( m_InterfaceDescriptor[ 0 ] ) ) == interfaceId )
    {
        if ( m_LoggerClient.OpenSection( LoggerClient::LOGGER_LEVEL__WARNING ) )
        {
            m_LoggerClient << STRING( "Can't lock detached reader [ S/N " ) << Manip( 10, ' ', 0 ) << UINT( serialNumber ) << STRING( " ]" ) << Endl;
            m_LoggerClient.Flush();
        }

        interfaceId = -1;
    }

    return interfaceId;
}

void ReaderFabric::UnlockInterface( int interfaceId )
{
    if ( ( interfaceId >= 0) && ( interfaceId < static_cast < int> ( sizeof ( m_InterfaceDescriptor ) / sizeof ( m_InterfaceDescriptor[ 0 ] ) ) ) )
    {
        ::pthread_mutex_lock( &m_InterfaceDescriptorLock );

        InterfaceDescriptor* pInterfaceDescriptor = &( m_InterfaceDescriptor[ interfaceId ] );
        if ( pInterfaceDescriptor->IsLocked )
        {
            pInterfaceDescriptor->IsLocked = false;

            if ( m_LoggerClient.OpenSection( LoggerClient::LOGGER_LEVEL__DEBUG ) )
            {
                m_LoggerClient << STRING( "Reader [ " );
                pInterfaceDescriptor->pReader->LogIdentity( m_LoggerClient );
                m_LoggerClient << STRING( " ] unlocked" ) << Endl;
                m_LoggerClient.Flush();
            }

            if ( INVALID_SERIAL_NUMBER == pInterfaceDescriptor->SerialNumber )
            {
                this->RemoveDevice( pInterfaceDescriptor );
            }
        }

        ::pthread_mutex_unlock( &m_InterfaceDescriptorLock );
    }
}

unsigned int ReaderFabric::GetAttachedReaders( unsigned long* pSerialNumberStorage, unsigned int serialNumberStorageEntries )
{
    unsigned int readerCount = 0;

    if ( ( 0 != pSerialNumberStorage ) && ( serialNumberStorageEntries > 0 ) )
    {
        ::pthread_mutex_lock( &m_InterfaceDescriptorLock );

        for ( unsigned int i = 0; i < sizeof ( m_InterfaceDescriptor ) / sizeof ( m_InterfaceDescriptor[ 0 ] ); i++ )
        {
            unsigned long serialNumber = m_InterfaceDescriptor[ i ].SerialNumber;
            if ( INVALID_SERIAL_NUMBER != serialNumber )
            {
                if ( readerCount < serialNumberStorageEntries )
                {
                    pSerialNumberStorage[ readerCount ] = serialNumber;
                    readerCount++;
                }
                else
                {
                    if ( m_LoggerClient.OpenSection( LoggerClient::LOGGER_LEVEL__WARNING ) )
                    {
                        m_LoggerClient << STRING( "Provided serial number storage size is not large enough" ) << Endl;
                        m_LoggerClient.Flush();
                    }

                    break;
                }
            }
        }

        ::pthread_mutex_unlock( &m_InterfaceDescriptorLock );
    }

    return readerCount;
}

void ReaderFabric::ReaderAdded( const char* pNode, unsigned short productId, unsigned long serialNumber )
{
    ::pthread_mutex_lock( &m_InterfaceDescriptorLock );

    InterfaceDescriptor* pInterfaceDescriptor = 0;
    for ( unsigned int i = 0; i < sizeof ( m_InterfaceDescriptor ) / sizeof ( m_InterfaceDescriptor[ 0 ] ); i++ )
    {
        InterfaceDescriptor* pCurrentInterfaceDescriptor = &( m_InterfaceDescriptor[ i ] );

        // Store the first available free slot
        if ( ( 0 == pInterfaceDescriptor ) && ( INVALID_SERIAL_NUMBER == pCurrentInterfaceDescriptor->SerialNumber ) )
        {
            pInterfaceDescriptor = pCurrentInterfaceDescriptor;
        }

        // Check if reader was already opened
        if ( ( pCurrentInterfaceDescriptor->ProductId == productId ) && ( pCurrentInterfaceDescriptor->SerialNumber == serialNumber ) )
        {
            pInterfaceDescriptor = 0;

            // Reader shall not be active
            if ( INVALID_SERIAL_NUMBER == pCurrentInterfaceDescriptor->SerialNumber )
            {
                pInterfaceDescriptor = pCurrentInterfaceDescriptor;
            }
            else if ( m_LoggerClient.OpenSection( LoggerClient::LOGGER_LEVEL__WARNING ) )
            {
                m_LoggerClient << STRING( "Can't add reader as it is already added" ) << Endl;
                m_LoggerClient.Flush();
            }

            break;
        }
    }

    if ( 0 != pInterfaceDescriptor )
    {
        switch ( productId )
        {
        case ReaderModelPrEh08::PRODUCT_ID:
            pInterfaceDescriptor->pReader = ReaderModelPrEh08::Create( pNode, serialNumber, m_ReaderLogger.pServer, m_ReaderLogger.Level );
            if ( 0 != pInterfaceDescriptor->pReader )
            {
                if ( m_LoggerClient.OpenSection( LoggerClient::LOGGER_LEVEL__INFO ) )
                {
                    m_LoggerClient << STRING( "PR-EH08 reader [ " );
                    pInterfaceDescriptor->pReader->LogIdentity( m_LoggerClient );
                    m_LoggerClient << STRING( " ]: created" ) << Endl;
                    m_LoggerClient.Flush();
                }
            }
            break;

        case ReaderModelPrp08::PRODUCT_ID:
            pInterfaceDescriptor->pReader = ReaderModelPrp08::Create( pNode, serialNumber, m_ReaderLogger.pServer, m_ReaderLogger.Level );
            if ( 0 != pInterfaceDescriptor->pReader )
            {
                if ( m_LoggerClient.OpenSection( LoggerClient::LOGGER_LEVEL__INFO ) )
                {
                    m_LoggerClient << STRING( "PRP-08 reader [ " );
                    pInterfaceDescriptor->pReader->LogIdentity( m_LoggerClient );
                    m_LoggerClient << STRING( " ]: created" ) << Endl;
                    m_LoggerClient.Flush();
                }
            }
            break;

        default:
            break;
        }

        if ( 0 != pInterfaceDescriptor->pReader )
        {
            pInterfaceDescriptor->ProductId = productId;
            pInterfaceDescriptor->SerialNumber = serialNumber;
        }
    }

    ::pthread_mutex_unlock( &m_InterfaceDescriptorLock );
}

void ReaderFabric::DeviceRemoved( const char* pNode )
{
    for ( unsigned int i = 0; i < sizeof ( m_InterfaceDescriptor ) / sizeof ( m_InterfaceDescriptor[ 0 ] ); i++ )
    {
        InterfaceDescriptor* pInterfaceDescriptor = &( m_InterfaceDescriptor[ i ] );

        if ( INVALID_SERIAL_NUMBER != pInterfaceDescriptor->SerialNumber )
        {
            if ( 0 == ::strcmp( pInterfaceDescriptor->pReader->GetNode(), pNode ) )
            {
                ::pthread_mutex_lock( &m_InterfaceDescriptorLock );

                if ( !pInterfaceDescriptor->IsLocked )
                {
                    this->RemoveDevice( pInterfaceDescriptor );
                }
                else
                {
                    if ( m_LoggerClient.OpenSection( LoggerClient::LOGGER_LEVEL__INFO ) )
                    {
                        m_LoggerClient << STRING( "Can't remove locked reader [ " );
                        pInterfaceDescriptor->pReader->LogIdentity( m_LoggerClient );
                        m_LoggerClient << STRING( " ]: postpone" ) << Endl;
                        m_LoggerClient.Flush();
                    }

                    pInterfaceDescriptor->SerialNumber = INVALID_SERIAL_NUMBER;
                }

                ::pthread_mutex_unlock( &m_InterfaceDescriptorLock );

                break;
            }
        }
    }
}

void ReaderFabric::RemoveDevice( InterfaceDescriptor* pInterfaceDescriptor )
{
    switch ( pInterfaceDescriptor->ProductId )
    {
    case ReaderModelPrEh08::PRODUCT_ID:
        if ( m_LoggerClient.OpenSection( LoggerClient::LOGGER_LEVEL__INFO ) )
        {
            m_LoggerClient << STRING( "PR-EH08 reader [ " );
            pInterfaceDescriptor->pReader->LogIdentity( m_LoggerClient );
            m_LoggerClient << STRING( " ]: removed" ) << Endl;
            m_LoggerClient.Flush();
        }
        ReaderModelPrEh08::Delete( reinterpret_cast < ReaderModelPrEh08* > ( pInterfaceDescriptor->pReader ) );
        break;

    case ReaderModelPrp08::PRODUCT_ID:
        if ( m_LoggerClient.OpenSection( LoggerClient::LOGGER_LEVEL__INFO ) )
        {
            m_LoggerClient << STRING( "PRP-08 reader [ " );
            pInterfaceDescriptor->pReader->LogIdentity( m_LoggerClient );
            m_LoggerClient << STRING( " ]: removed" ) << Endl;
            m_LoggerClient.Flush();
        }
        ReaderModelPrp08::Delete( reinterpret_cast < ReaderModelPrp08* > ( pInterfaceDescriptor->pReader ) );
        break;

    default:
        break;
    }

    ::memset( pInterfaceDescriptor, 0, sizeof ( InterfaceDescriptor ) );
}
