// Standard includes
#include <unistd.h>
#include <string.h>

// Vendor includes

// Project includes
#include "LoggerServerImpl.hpp"
#include "ReaderFabric.hpp"
#include "reader_interface.h"

// Using declarations


static LoggerServerImpl loggerServerImpl;
static ReaderFabric readerFabric;
static bool isWiegand26 = false;


static LoggerClient::LoggerLevel TranslateDebugLevel( int debugLevel )
{
    LoggerClient::LoggerLevel loggerLevel;

    switch ( debugLevel )
    {
    case 0:
        loggerLevel = LoggerClient::LOGGER_LEVEL__ERROR;
        break;

    case 2:
        loggerLevel = LoggerClient::LOGGER_LEVEL__INFO;
        break;

    case 3:
        loggerLevel = LoggerClient::LOGGER_LEVEL__DEBUG;
        break;

    default:
        loggerLevel = LoggerClient::LOGGER_LEVEL__WARNING;
        break;
    }

    return loggerLevel;
}

void reader_set_log_func( debug_log_func_t debug_log_func_ptr )
{
    loggerServerImpl.SetExternalLogger( debug_log_func_ptr );
}

void reader_debug_level( int fabricDebugLevel, int plugAndPlayDebugLevel, int readerObjectDebugLevel )
{
    readerFabric.SetLogger( &loggerServerImpl, TranslateDebugLevel( fabricDebugLevel ), TranslateDebugLevel( plugAndPlayDebugLevel ) );
    readerFabric.SetReaderLogger( &loggerServerImpl, TranslateDebugLevel( readerObjectDebugLevel ) );
}

int reader_interface_init()
{
    int retCode = -1;

    if ( loggerServerImpl.ThreadStart( 0 ) )
    {
        if ( loggerServerImpl.Open() )
        {
            if ( readerFabric.Initialize() )
            {
                retCode = 0;
            }
        }
    }

    return retCode;
}

void reader_interface_term()
{
    readerFabric.Terminate();

    loggerServerImpl.ScheduleClose();
    while ( loggerServerImpl.IsRunning() )
    {
        ::usleep( 100000 );
    }

    loggerServerImpl.ThreadStop();
}

unsigned int reader_get_attached( unsigned long* serial_number_storage_ptr, unsigned int serial_number_storage_entries )
{
    return readerFabric.GetAttachedReaders( serial_number_storage_ptr, serial_number_storage_entries );
}

int reader_get_info( unsigned long serialNumber, Get_info_response_t* pGeInfoResponse )
{
    int retCode = -1;

    if ( 0 != pGeInfoResponse )
    {
        unsigned short productId;
        ReaderBase* pReader;
        int interfaceId = readerFabric.LockInterface( serialNumber, productId, pReader );
        if ( interfaceId >= 0 )
        {
            switch ( productId )
            {
            case ReaderModelPrEh08::PRODUCT_ID:
            {
                ReaderModelPrEh08::Response::GetInfo responseGetInfo;
                if ( ( reinterpret_cast < ReaderModelPrEh08* > ( pReader )->ProcessGetInfo( responseGetInfo ) ) &&
                     ( ReaderBase::ResponseHeader::RESULT__RES_OK == responseGetInfo.Header.Result ) &&
                     ( ( sizeof ( ReaderModelPrEh08::Response::GetInfo ) - sizeof ( ReaderModelPrEh08::Response::GetInfo::Header ) )
                       == responseGetInfo.Header.Length ) )
                {
                    ::memcpy( &( pGeInfoResponse->product_info[ 0 ] ), &( responseGetInfo.ProductInfo[ 0 ] ),
                              sizeof ( pGeInfoResponse->product_info ) );
                    ::memcpy( &( pGeInfoResponse->chip_serial_number[ 0 ] ), &( responseGetInfo.ChipSerialNumber[ 0 ] ),
                              sizeof ( pGeInfoResponse->chip_serial_number ) );
                    pGeInfoResponse->version_ascii.one = responseGetInfo.VersionAscii.One;
                    pGeInfoResponse->version_ascii.two = responseGetInfo.VersionAscii.Two;
                    pGeInfoResponse->supported_interfaces.all = responseGetInfo.SupportedInterfaces.All;
                    pGeInfoResponse->processor_type = static_cast < Processor_type > ( responseGetInfo.ProcessorType );
                    pGeInfoResponse->reader_mode = static_cast < Reader_mode > ( responseGetInfo.ReaderMode );
                    pGeInfoResponse->rf_state = static_cast < Rf_state > ( responseGetInfo.RfState );
                    pGeInfoResponse->reader_address = responseGetInfo.ReaderAddress;

                    retCode = 0;
                }
            }
                break;

            case ReaderModelPrp08::PRODUCT_ID:
                {
                    ReaderModelPrp08::Response::GetInfo responseGetInfo;
                    if ( ( reinterpret_cast < ReaderModelPrp08* > ( pReader )->ProcessGetInfo( responseGetInfo ) ) &&
                         ( ReaderBase::ResponseHeader::RESULT__RES_OK == responseGetInfo.Header.Result ) &&
                         ( ( sizeof ( ReaderModelPrp08::Response::GetInfo ) - sizeof ( ReaderModelPrp08::Response::GetInfo::Header ) )
                           == responseGetInfo.Header.Length ) )
                    {
                        ::memcpy( &( pGeInfoResponse->product_info[ 0 ] ), &( responseGetInfo.ProductInfo[ 0 ] ),
                                  sizeof ( pGeInfoResponse->product_info ) );
                        ::memcpy( &( pGeInfoResponse->chip_serial_number[ 0 ] ), &( responseGetInfo.ChipSerialNumber[ 0 ] ),
                                  sizeof ( pGeInfoResponse->chip_serial_number ) );
                        pGeInfoResponse->version_ascii.one = responseGetInfo.VersionAscii.One;
                        pGeInfoResponse->version_ascii.two = responseGetInfo.VersionAscii.Two;
                        pGeInfoResponse->supported_interfaces.all = responseGetInfo.SupportedInterfaces.All;
                        pGeInfoResponse->processor_type = static_cast < Processor_type > ( responseGetInfo.ProcessorType );
                        pGeInfoResponse->reader_mode = static_cast < Reader_mode > ( responseGetInfo.ReaderMode );
                        pGeInfoResponse->rf_state = static_cast < Rf_state > ( responseGetInfo.RfState );
                        pGeInfoResponse->reader_address = responseGetInfo.ReaderAddress;

                        retCode = 0;
                    }
                }
                break;

            default:
                break;
            }

            readerFabric.UnlockInterface( interfaceId );
        }
    }

    return retCode;
}

int reader_set_beep_and_blink_mode( unsigned long serialNumber, unsigned char beep_on, unsigned char blink_on )
{
    int retCode = -1;

    unsigned short productId;
    ReaderBase* pReader;
    int interfaceId = readerFabric.LockInterface( serialNumber, productId, pReader );
    if ( interfaceId >= 0 )
    {
        switch ( productId )
        {
        case ReaderModelPrEh08::PRODUCT_ID:
            {
                ReaderModelPrEh08::Response::SetBeepAndBlink responseSetBeepAndBlink;
                if ( ( reinterpret_cast < ReaderModelPrEh08* > ( pReader )->ProcessSetBeepAndBlink( beep_on, blink_on, responseSetBeepAndBlink ) ) &&
                     ( ReaderBase::ResponseHeader::RESULT__RES_OK == responseSetBeepAndBlink.Header.Result ) &&
                     ( 0 == responseSetBeepAndBlink.Header.Length ) )
                {
                    retCode = 0;
                }
            }
            break;

        case ReaderModelPrp08::PRODUCT_ID:
            {
                ReaderModelPrp08::Response::SetBeepAndBlink responseSetBeepAndBlink;
                if ( ( reinterpret_cast < ReaderModelPrp08* > ( pReader )->ProcessSetBeepAndBlink( beep_on, blink_on, responseSetBeepAndBlink ) ) &&
                     ( ReaderBase::ResponseHeader::RESULT__RES_OK == responseSetBeepAndBlink.Header.Result ) &&
                     ( 0 == responseSetBeepAndBlink.Header.Length ) )
                {
                    retCode = 0;
                }
            }
            break;

        default:
            break;
        }

        readerFabric.UnlockInterface( interfaceId );
    }

    return retCode;
}

int reader_beep( unsigned long serialNumber )
{
    int retCode = -1;

    unsigned short productId;
    ReaderBase* pReader;
    int interfaceId = readerFabric.LockInterface( serialNumber, productId, pReader );
    if ( interfaceId >= 0 )
    {
        switch ( productId )
        {
        case ReaderModelPrEh08::PRODUCT_ID:
            {
                ReaderModelPrEh08::Response::Beep responseBeep;
                if ( ( reinterpret_cast < ReaderModelPrEh08* > ( pReader )->ProcessBeep( responseBeep ) ) &&
                     ( ReaderBase::ResponseHeader::RESULT__RES_OK == responseBeep.Header.Result ) &&
                     ( 0 == responseBeep.Header.Length ) )
                {
                    retCode = 0;
                }
            }
            break;

        case ReaderModelPrp08::PRODUCT_ID:
            {
                ReaderModelPrp08::Response::Beep responseBeep;
                if ( ( reinterpret_cast < ReaderModelPrp08* > ( pReader )->ProcessBeep( responseBeep ) ) &&
                     ( ReaderBase::ResponseHeader::RESULT__RES_OK == responseBeep.Header.Result ) &&
                     ( 0 == responseBeep.Header.Length ) )
                {
                    retCode = 0;
                }
            }
            break;

        default:
            break;
        }

        readerFabric.UnlockInterface( interfaceId );
    }

    return retCode;
}

void reader_set_wiegand_26( unsigned char wiegand_26_on )
{
    isWiegand26 = wiegand_26_on;
}

int reader_read_card( unsigned long serialNumber, unsigned long* pCardNumber, Read_card_response_t* pReadCardResponse )
{
    int retCode = -1;

    if ( 0 != pReadCardResponse )
    {
        unsigned short productId;
        ReaderBase* pReader;
        int interfaceId = readerFabric.LockInterface( serialNumber, productId, pReader );
        if ( interfaceId >= 0 )
        {
            switch ( productId )
            {
            case ReaderModelPrEh08::PRODUCT_ID:
                {
                    ReaderModelPrEh08::Response::ReadCard responseReadCard;
                    if ( ( reinterpret_cast < ReaderModelPrEh08* > ( pReader )->ProcessReadCard( responseReadCard ) ) &&
                         ( ReaderBase::ResponseHeader::RESULT__RES_OK == responseReadCard.Header.Result ) )
                    {
                        retCode = 0;
                        switch ( responseReadCard.CardType )
                        {
                        case ReaderModelPrEh08::Response::ReadCard::CARD_TYPE__EM_MARINE:
                            if ( 0 != pCardNumber )
                            {
                                *pCardNumber = ( responseReadCard.Data[ 1 ] << 24 ) | ( responseReadCard.Data[ 2 ] << 16 ) |
                                               ( responseReadCard.Data[ 3 ] << 8 ) | responseReadCard.Data[ 4 ];
                                if ( isWiegand26 )
                                {
                                    *pCardNumber &= 0x00FFFFFF;
                                }
                            }
                            if ( 0 != pReadCardResponse )
                            {
                                pReadCardResponse->card_type = CARD_TYPE__EM_MARINE;
                            }
                            break;

                        case ReaderModelPrEh08::Response::ReadCard::CARD_TYPE__HID:
                            if ( 0 != pCardNumber )
                            {
                                *pCardNumber = ( responseReadCard.Data[ 2 ] << 24 ) | ( responseReadCard.Data[ 3 ] << 16 ) |
                                               ( responseReadCard.Data[ 4 ] << 8 ) | responseReadCard.Data[ 5 ];
                                if ( isWiegand26 )
                                {
                                    if ( 0 == responseReadCard.Data[ 0 ] )
                                    {
                                        *pCardNumber <<= 1;
                                    }
                                    *pCardNumber = ( *pCardNumber & 0x0FFFFFF0 ) >> 4;
                                }
                            }
                            if ( 0 != pReadCardResponse )
                            {
                                pReadCardResponse->card_type = CARD_TYPE__HID;
                            }
                            break;

                        default:
                            retCode = -1;
                            break;
                        }

                        if ( 0 != pReadCardResponse )
                        {
                            pReadCardResponse->card.em_marine_hid.length = responseReadCard.Header.Length - sizeof ( responseReadCard.CardType );
                            if ( sizeof ( Read_card_response_t::card.em_marine_hid.data ) >= pReadCardResponse->card.em_marine_hid.length )
                            {
                                ::memcpy( &( pReadCardResponse->card.em_marine_hid.data[ 0 ] ), &( responseReadCard.Data[ 0 ] ),
                                          pReadCardResponse->card.em_marine_hid.length );
                            }
                        }
                    }
                }
                break;

            case ReaderModelPrp08::PRODUCT_ID:
                {
                    ReaderModelPrp08::Response::Rf responseRf;
                    ReaderModelPrp08::Response::ReadCard responseReadCard;
                    ReaderModelPrp08* pReaderModelPrp08 = reinterpret_cast < ReaderModelPrp08* > ( pReader );
                    ::memset( &( responseReadCard.Uid[ 0 ] ), 0, sizeof ( responseReadCard.Uid ) );
                    if ( pReaderModelPrp08->ProcessRf( ReaderModelPrp08::Command::Rf::CARD_TYPE__MIFARE, ReaderModelPrp08::Command::Rf::RF_STATE__HIGH_POWER,
                             responseRf ) && ( ReaderBase::ResponseHeader::RESULT__RES_OK == responseRf.Header.Result ) &&
                         pReaderModelPrp08->ProcessReadCard( ReaderModelPrp08::Command::ReadCard::REQUEST_TYPE__HALTED_CARD_ACTIVATE, responseReadCard ) &&
                             ( ReaderBase::ResponseHeader::RESULT__RES_OK == responseReadCard.Header.Result ) )
                    {
                        if ( 0 != pCardNumber )
                        {
                            *pCardNumber = ( responseReadCard.Uid[ 0 ] << 24 ) | ( responseReadCard.Uid[ 1 ] << 16 ) |
                                           ( responseReadCard.Uid[ 2 ] << 8 ) | responseReadCard.Uid[ 3 ];
                            if ( isWiegand26 )
                            {
                                *pCardNumber &= 0x00FFFFFF;
                            }
                        }

                        if ( 0 != pReadCardResponse )
                        {
                            pReadCardResponse->card_type = CARD_TYPE__MIFARE;
                            pReadCardResponse->card.mifare.atqa.low = responseReadCard.Atqa.Low;
                            pReadCardResponse->card.mifare.atqa.high = responseReadCard.Atqa.High;
                            pReadCardResponse->card.mifare.sak = responseReadCard.Sak;
                            pReadCardResponse->card.mifare.uid_length = ( responseReadCard.UidLength <= sizeof ( responseReadCard.Uid ) ) ?
                                responseReadCard.UidLength : sizeof ( responseReadCard.Uid );
                            ::memcpy( &( pReadCardResponse->card.mifare.uid[ 0 ] ), &( responseReadCard.Uid[ 0 ] ), pReadCardResponse->card.mifare.uid_length );
                        }

                        retCode = 0;
                    }

                    // Switch off the RF - do not care about the result
                    pReaderModelPrp08->ProcessRf( ReaderModelPrp08::Command::Rf::CARD_TYPE__NO_CHANGE, ReaderModelPrp08::Command::Rf::RF_STATE__OFF, responseRf );
                }
                break;

            default:
                break;
            }

            readerFabric.UnlockInterface( interfaceId );
        }
    }

    return retCode;
}
