// Standard includes
#include <stdlib.h>
#include <unistd.h>
#include <sys/poll.h>
#include <string.h>

// Vendor includes

// Project includes
#include "LoggerServerImpl.hpp"
#include "ReaderPlugAndPlay.hpp"
#include "ReaderModelPrEh08.hpp"
#include "ReaderModelPrp08.hpp"

// Using declarations


ReaderPlugAndPlay::ReaderPlugAndPlay()
    : m_LoggerClient( LoggerServerImpl::LOGGER_CLIENT_ID__READER_PLUG_AND_PLAY ),
      m_pUdevContext( 0 ),
      m_pUdevMonitor( 0 ),
      m_IsRunning( false )
{
}

ReaderPlugAndPlay::~ReaderPlugAndPlay()
{
    if ( m_IsRunning )
    {
        this->ScheduleClose();

        while ( m_IsRunning )
        {
            // Need to wait until thread exits
            ::usleep( 10000 );
        }
        this->ThreadStop();
    }
}

void ReaderPlugAndPlay::SetLogger( LoggerServer* pLoggerServer, LoggerClient::LoggerLevel loggerLevel )
{
    if ( 0 != pLoggerServer )
    {
        m_LoggerClient.AttachLoggerServer( pLoggerServer );
    }
    m_LoggerClient.SetLevel( loggerLevel );
}

bool ReaderPlugAndPlay::Open()
{
    if ( !m_IsRunning )
    {
        // Create device manager context
        m_pUdevContext = ::udev_new();
        if ( 0 != m_pUdevContext )
        {
            // Create and setup device manager monitor
            m_pUdevMonitor = ::udev_monitor_new_from_netlink( m_pUdevContext, "udev" );
            if ( 0 != m_pUdevMonitor )
            {
                ::udev_monitor_filter_add_match_subsystem_devtype( m_pUdevMonitor, "tty", 0 );
                ::udev_monitor_enable_receiving( m_pUdevMonitor );

                if ( 0 == ::pipe( &( m_Pipe[ 0 ] ) ) )
                {
                    if ( this->ThreadActivate() )
                    {
                        if ( m_LoggerClient.OpenSection( LoggerClient::LOGGER_LEVEL__INFO ) )
                        {
                            m_LoggerClient << STRING( "Device manager opened" ) << Endl;
                            m_LoggerClient.Flush();
                        }

                        m_IsRunning = true;
                    }
                    else
                    {
                        if ( m_LoggerClient.OpenSection( LoggerClient::LOGGER_LEVEL__ERROR ) )
                        {
                            m_LoggerClient << STRING( "Can't activate device manager monitor" ) << Endl;
                            m_LoggerClient.Flush();
                        }
                    }
                }
                else if ( m_LoggerClient.OpenSection( LoggerClient::LOGGER_LEVEL__ERROR ) )
                {
                    m_LoggerClient << STRING( "Can't create device manager control queue" ) << Endl;
                    m_LoggerClient.Flush();
                }
            }
            else if ( m_LoggerClient.OpenSection( LoggerClient::LOGGER_LEVEL__ERROR ) )
            {
                m_LoggerClient << STRING( "Can't create device manager monitor" ) << Endl;
                m_LoggerClient.Flush();
            }
        }
        else if ( m_LoggerClient.OpenSection( LoggerClient::LOGGER_LEVEL__ERROR ) )
        {
            m_LoggerClient << STRING( "Can't create device manager context" ) << Endl;
            m_LoggerClient.Flush();
        }

        if ( !m_IsRunning )
        {
            this->ReleaseResorces();
        }
    }
    else if ( m_LoggerClient.OpenSection( LoggerClient::LOGGER_LEVEL__WARNING ) )
    {
        m_LoggerClient << STRING( "Device manager already running" ) << Endl;
        m_LoggerClient.Flush();
    }

    return m_IsRunning;
}

void ReaderPlugAndPlay::ScheduleClose()
{
    // Unblock the receive - do not care about the data
    ::write( m_Pipe[ PIPE_CHANNEL__WRITE ], &( m_Pipe[ PIPE_CHANNEL__WRITE ] ), sizeof ( m_Pipe[ PIPE_CHANNEL__WRITE ] ) );

    if ( m_LoggerClient.OpenSection( LoggerClient::LOGGER_LEVEL__INFO ) )
    {
        m_LoggerClient << STRING( "Device manager termination scheduled" ) << Endl;
        m_LoggerClient.Flush();
    }
}

bool ReaderPlugAndPlay::IsRunning() const
{
    return m_IsRunning;
}

void ReaderPlugAndPlay::Run( void* pUserData )
{
    ::pollfd pollFd[ 2 ];
    pollFd[ 0 ].fd = m_Pipe[ PIPE_CHANNEL__READ ];
    pollFd[ 0 ].events = POLLIN;
    pollFd[ 1 ].fd = ::udev_monitor_get_fd( m_pUdevMonitor );
    pollFd[ 1 ].events = POLLIN;

    this->ScanBus();

    while ( true )
    {
        // Wait for system events
        if ( ::poll( &( pollFd[ 0 ] ), 2, -1 ) > 0 )
        {
            if ( pollFd[ 1 ].revents & POLLIN )
            {
                this->ProcessTtyDevice( ::udev_monitor_receive_device( m_pUdevMonitor ) );
            }

            if ( pollFd[ 0 ].revents & POLLIN )
            {
                // Close scheduled
                break;
            }
        }
    }

    this->ReleaseResorces();
    m_IsRunning = false;

    if ( m_LoggerClient.OpenSection( LoggerClient::LOGGER_LEVEL__INFO ) )
    {
        m_LoggerClient << STRING( "Device manager closed" ) << Endl;
        m_LoggerClient.Flush();
    }
}

void ReaderPlugAndPlay::ReleaseResorces()
{
    if ( m_pUdevMonitor != 0 )
    {
        // Destroy all filters on monitor
        ::udev_monitor_filter_remove( m_pUdevMonitor );

        // Destroy device manager monitor
        ::udev_monitor_unref( m_pUdevMonitor );
    }

    // Destroy device manager
    ::udev_unref( m_pUdevContext );

    ::close( m_Pipe[ PIPE_CHANNEL__READ ] );
    ::close( m_Pipe[ PIPE_CHANNEL__WRITE ] );
}

void ReaderPlugAndPlay::ScanBus()
{
    if ( 0 != m_pUdevContext )
    {
        ::udev_enumerate* pEnumerateContext = ::udev_enumerate_new( m_pUdevContext );
        if ( 0 != pEnumerateContext )
        {
            if ( ( 0 == ::udev_enumerate_add_match_subsystem( pEnumerateContext, "tty" ) ) && ( 0 == ::udev_enumerate_scan_devices( pEnumerateContext ) ) )
            {
                ::udev_list_entry* pDeviceListEntry = udev_enumerate_get_list_entry( pEnumerateContext );
                while ( 0 != pDeviceListEntry )
                {
                    // Process TTY device found
                    this->ProcessTtyDevice( ::udev_device_new_from_syspath( m_pUdevContext, ::udev_list_entry_get_name( pDeviceListEntry ) ) );

                    pDeviceListEntry = ::udev_list_entry_get_next( pDeviceListEntry );
                }
            }
            else if ( m_LoggerClient.OpenSection( LoggerClient::LOGGER_LEVEL__WARNING ) )
            {
                m_LoggerClient << STRING( "Can't scan the bus" ) << Endl;
                m_LoggerClient.Flush();
            }

            ::udev_enumerate_unref( pEnumerateContext );
        }
        else if ( m_LoggerClient.OpenSection( LoggerClient::LOGGER_LEVEL__WARNING ) )
        {
            m_LoggerClient << STRING( "Can't create enumeration context" ) << Endl;
            m_LoggerClient.Flush();
        }
    }
}

void ReaderPlugAndPlay::ProcessTtyDevice( ::udev_device* pDevice )
{
    if ( 0 != pDevice )
    {
        const char* pNode = ::udev_device_get_devnode( pDevice );
        const char* pAction = ::udev_device_get_action( pDevice );

        if ( ( 0 == pAction ) || ( 0 == ::strcmp( pAction, "add" ) ) )
        {
            // TTY device added
            ::udev_device* pParentDevice = ::udev_device_get_parent_with_subsystem_devtype( pDevice, "usb", "usb_device" );

            if ( ( 0 != pParentDevice ) && ( 0 == ::strcmp( ::udev_device_get_sysattr_value( pParentDevice, "idVendor" ), "0403" ) ) )
            {
                long attrProductId = ::strtol( ::udev_device_get_sysattr_value( pParentDevice, "idProduct" ), 0, 16 );
                switch ( attrProductId )
                {
                case ReaderModelPrEh08::PRODUCT_ID:
                case ReaderModelPrp08::PRODUCT_ID:
                    {
                        unsigned short productId;
                        unsigned long serialNumber;
                        if ( this->DecodeSerialUtf8( ::udev_device_get_sysattr_value( pParentDevice, "serial" ), productId, serialNumber ) )
                        {
                            if ( attrProductId == productId )
                            {
                                if ( m_LoggerClient.OpenSection( LoggerClient::LOGGER_LEVEL__INFO ) )
                                {
                                    m_LoggerClient << STRING( "Reader [ " ) << STRING( pNode ) << STRING( ", S/N " ) << Manip( 10, ' ', 0 )
                                                   << UINT( serialNumber ) << STRING( ", PID 0x" ) << Manip( 16, '0', 4 ) << UINT( productId )
                                                   << STRING( " ]: attached" ) << Endl;
                                    m_LoggerClient.Flush();
                                }

                                // Reader add notification
                                this->ReaderAdded( pNode, productId, serialNumber );
                            }
                            else if ( m_LoggerClient.OpenSection( LoggerClient::LOGGER_LEVEL__WARNING ) )
                            {
                                m_LoggerClient << STRING( "Reader [ " ) << STRING( pNode ) << STRING( ", S/N " ) << Manip( 10, ' ', 0 )
                                               << UINT( serialNumber ) << STRING( " ]: PID mismatch - decoded 0x" ) << Manip( 16, '0', 4 ) << UINT( productId )
                                               << STRING( ", expected 0x)" ) << UINT( attrProductId ) << Endl;
                                m_LoggerClient.Flush();
                            }
                        }
                        else if ( m_LoggerClient.OpenSection( LoggerClient::LOGGER_LEVEL__WARNING ) )
                        {
                            m_LoggerClient << STRING( "Reader [ " ) << STRING( pNode ) << STRING( " ]: serial number is not valid" ) << Endl;
                            m_LoggerClient.Flush();
                        }
                    }
                    break;

                default:
                    // Some other FTDI device - do nothing
                    break;
                }
            }
        }
        else if ( 0 == ::strcmp( pAction, "remove" ) )
        {
            if ( m_LoggerClient.OpenSection( LoggerClient::LOGGER_LEVEL__INFO ) )
            {
                m_LoggerClient << STRING( "Reader [ " ) << STRING( pNode ) << STRING( " ]: detached" ) << Endl;
                m_LoggerClient.Flush();
            }

            // TTY device remove notification
            this->DeviceRemoved( pNode );
        }

        ::udev_device_unref( pDevice );
    }
}

bool ReaderPlugAndPlay::DecodeSerialUtf8( const char* pSerialUtf8, unsigned short& rProductId, unsigned long& rSerialNumber )
{
    bool isSuccess = false;

    // This method does not use any constants because serial number translation algorithm is hard to formalize

    // Get coded serial number from UTF8 string
    if ( 0 != pSerialUtf8 )
    {
        // Coded serial number consists of 8 bytes
        unsigned char codedSerialNumber[ 8 ];

        unsigned int utf8SymbolBytesLeft = 0;
        unsigned int dataLength = 0;
        while ( ( *pSerialUtf8 != 0 ) && ( dataLength < sizeof ( codedSerialNumber ) ) )
        {
            if ( 0 == utf8SymbolBytesLeft )
            {
                if ( *pSerialUtf8 & 0x80 )
                {
                    if ( 0xC0 == ( *pSerialUtf8 & 0xE0 ) )
                    {
                        // 2-byte symbol, 1 byte result
                        utf8SymbolBytesLeft = 1;
                        codedSerialNumber[ dataLength ] = *pSerialUtf8 & 0x1F;
                    }
                    else if ( 0xE0 == ( *pSerialUtf8 & 0xF0 ) )
                    {
                        // 3-byte symbol, 2 byte result ( high byte of a 2 byte result is truncated as it is not required [ 0xFF ] )
                        utf8SymbolBytesLeft = 2;
                        codedSerialNumber[ dataLength ] = *pSerialUtf8 & 0x0F;
                    }
                    else
                    {
                        // Longer symbols are not supported
                        break;
                    }
                }
                else
                {
                    codedSerialNumber[ dataLength ] = *pSerialUtf8;
                    dataLength++;
                }
            }
            else if ( 0x80 == ( *pSerialUtf8 & 0xC0 ) )
            {
                codedSerialNumber[ dataLength ] <<= 6;
                codedSerialNumber[ dataLength ] |= *pSerialUtf8 & 0x3F;
                if ( 0 == --utf8SymbolBytesLeft )
                {
                    dataLength++;
                }
            }
            else
            {
                // UTF8 error
                break;
            }

            pSerialUtf8++;
        }

        // Proceed in case of success
        if ( ( 0 == *pSerialUtf8 ) && ( dataLength == sizeof ( codedSerialNumber ) ) )
        {
            rProductId = ( static_cast < unsigned short > ( codedSerialNumber[ 0 ] ) << 8 ) | codedSerialNumber[ 1 ];

            if ( codedSerialNumber[ 2 ] > 0 )
            {
                rSerialNumber = codedSerialNumber[ 2 ] - 1;

                // Decode BCD numbers - reuse original buffer
                isSuccess = true;
                for ( unsigned int i = 5; isSuccess && ( i < sizeof ( codedSerialNumber ) ); i++ )
                {
                    unsigned char bcdByte = codedSerialNumber[ i ];
                    if ( ( ( bcdByte & 0xF0 ) <= 0xA0 ) && ( ( bcdByte & 0xF0 ) >= 0x10 ) && ( ( bcdByte & 0x0F ) <= 0x0A ) && ( ( bcdByte & 0x0F ) >= 0x01 ) )
                    {
                        rSerialNumber = ( rSerialNumber * 100 ) + ( ( bcdByte >> 4 ) - 1 ) * 10 + ( ( bcdByte & 0x0F ) - 1 );
                    }
                    else
                    {
                        isSuccess = false;
                    }
                }

                if ( isSuccess )
                {
                    switch ( rProductId )
                    {
                    case ReaderModelPrp08::PRODUCT_ID:
                        rSerialNumber += 872010000; // Fixed base error by adding 10000
                        break;

                    case ReaderModelPrEh08::PRODUCT_ID:
                        rSerialNumber += 820000000;
                        break;

                    default:
                        isSuccess = false;
                        break;
                    }
                }
            }
        }
    }

    return isSuccess;
}
