// Standard includes
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>

// Vendor includes

// Project includes
#include "UsbSerial.hpp"

// Using declarations


const UsbSerial::SpeedMapRecord UsbSerial::s_SpeedMap[] =
{
        // Speed        Speed code
        {  9600,        B9600   },
        {  19200,       B19200  },
        {  38400,       B38400  },
        {  57600,       B57600  },
        {  115200,      B115200 },
        {  230400,      B230400 },
        {  460800,      B460800 }
};

UsbSerial::UsbSerial()
    : m_LastError( USB_SERIAL_ERROR__SUCCESS )
{
    m_PollFd.fd = -1;
    m_PollFd.events = POLLIN;
}

UsbSerial::~UsbSerial()
{
    this->Close();
}

bool UsbSerial::Open( const char* pNode, unsigned int speed, char parityCode, unsigned char stopBits )
{
    m_LastError = USB_SERIAL_ERROR__DEVICE_ALREADY_OPENED;

    if ( m_PollFd.fd < 0 )
    {
        m_LastError = USB_SERIAL_ERROR__PARAMETER_ERROR;

        if ( 0 != pNode )
        {
            // Map provided speed value to speed code
            unsigned int speedCode = 0;
            for ( unsigned int i = 0; i < ( sizeof ( s_SpeedMap ) / sizeof ( s_SpeedMap[ 0 ] ) ); i++ )
            {
                if ( s_SpeedMap[ i ].Speed == speed )
                {
                    speedCode = s_SpeedMap[ i ].SpeedCode;
                    break;
                }
            }

            // Check provided port parameters
            if ( ( 0 != speedCode ) && ( ( 'N' == parityCode ) || ( 'E' == parityCode ) || ( 'O' == parityCode ) ) &&
                 ( ( 1 == stopBits ) || ( 2 == stopBits ) ) )
            {
                // Check if specified USB device exists
                m_LastError = USB_SERIAL_ERROR__DEVICE_NOT_FOUND;
                if ( -1 != ::access( pNode, F_OK ) )
                {
                    // Open application UART for read/write operations,
                    // disable process control and set synchronous I/O mode
                    m_PollFd.fd = ::open( pNode, O_RDWR | O_NOCTTY | O_SYNC );
                    m_LastError = USB_SERIAL_ERROR__DEVICE_OPEN_ERROR;
                    if ( m_PollFd.fd >= 0 )
                    {
                        m_LastError = USB_SERIAL_ERROR__DEVICE_IO_OPERATION_ERROR;

                        // Get current terminal control parameters
                        if ( 0 == ::tcgetattr( m_PollFd.fd, &m_TtyControl ) )
                        {
                            // Interval to wait after the last symbol is received ( in 100 milliseconds )
                            m_TtyControl.c_cc[ VTIME ] = 1;
                            // Input modes
                            m_TtyControl.c_iflag = 0;
                            // Output modes
                            m_TtyControl.c_oflag = 0;
                            // Local modes
                            m_TtyControl.c_lflag = 0;
                            // Control modes:
                            m_TtyControl.c_cflag = 0;
                            // - enable receiver
                            m_TtyControl.c_cflag |= ( CREAD | CLOCAL );
                            // - set 8 bit symbol width
                            m_TtyControl.c_cflag |= CS8;
                            // - set parity
                            switch ( parityCode )
                            {
                            case 'E':
                                m_TtyControl.c_cflag |= PARENB;
                                break;
                            case 'O':
                                m_TtyControl.c_cflag |= ( PARENB | PARODD );
                                break;
                            }
                            // - set the number of stop bits
                            if ( 2 == stopBits )
                            {
                                m_TtyControl.c_cflag |= CSTOPB;
                            }

                            if ( ( 0 == ::cfsetospeed( &m_TtyControl, speedCode ) ) &&
                                 ( 0 == ::cfsetispeed( &m_TtyControl, speedCode ) ) &&
                                 ( 0 == ::tcsetattr( m_PollFd.fd, TCSANOW, &m_TtyControl ) ) &&
                                 ( 0 == ::tcflush( m_PollFd.fd, TCIOFLUSH ) ) )
                            {
                                m_LastError = USB_SERIAL_ERROR__SUCCESS;
                            }
                        }
                    }
                }
            }
        }
    }

    return ( USB_SERIAL_ERROR__SUCCESS == m_LastError );
}

void UsbSerial::Close()
{
    // Reset last error
    m_LastError = USB_SERIAL_ERROR__SUCCESS;

    if ( m_PollFd.fd >= 0 )
    {
        ::close( m_PollFd.fd );

        m_PollFd.fd = -1;
    }
}

bool UsbSerial::Read( unsigned char* pBuffer, unsigned int expectedLength, int timeToWaitInMilliseconds )
{
    m_LastError = USB_SERIAL_ERROR__PARAMETER_ERROR;

    if ( ( 0 != pBuffer ) && ( 0 != expectedLength ) )
    {
        m_LastError = USB_SERIAL_ERROR__DEVICE_NOT_OPENED;

        if ( m_PollFd.fd >= 0 )
        {
            m_TtyControl.c_cc[ VMIN ] = ( expectedLength < 256 ) ? expectedLength : 255;

            m_LastError = USB_SERIAL_ERROR__DEVICE_IO_OPERATION_ERROR;
            if ( 0 == ::tcsetattr( m_PollFd.fd, TCSANOW, &m_TtyControl ) )
            {
                m_LastError = USB_SERIAL_ERROR__SUCCESS;

                // Read all requested bytes
                unsigned int bytesToRead = expectedLength;
                int bytesRead;
                while ( bytesToRead > 0 )
                {
                    int pollResult = ::poll( &m_PollFd, 1, timeToWaitInMilliseconds );
                    if ( pollResult > 0 )
                    {
                        if ( m_PollFd.revents & POLLHUP )
                        {
                            m_LastError = USB_SERIAL_ERROR__DEVICE_NOT_OPENED;
                            break;
                        }
                        else if ( m_PollFd.revents & POLLIN )
                        {
                            bytesRead = ::read( m_PollFd.fd, pBuffer + expectedLength - bytesToRead, bytesToRead );
                            if ( bytesRead < 0 )
                            {
                                // Ignore error if blocking operation was interrupted by a signal
                                if ( EINTR == errno )
                                {
                                    continue;
                                }

                                m_LastError = USB_SERIAL_ERROR__DEVICE_IO_OPERATION_ERROR;
                                break;
                            }
                            bytesToRead -= bytesRead;
                        }
                    }
                    else if ( 0 == pollResult )
                    {
                        m_LastError = USB_SERIAL_ERROR__DEVICE_IO_OPERATION_TIMEOUT;
                        break;
                    }
                }
            }
        }
    }

    return ( USB_SERIAL_ERROR__SUCCESS == m_LastError );
}

bool UsbSerial::Write( unsigned char* pBuffer, unsigned int length )
{
    m_LastError = USB_SERIAL_ERROR__PARAMETER_ERROR;

    if ( ( 0 != pBuffer ) && ( 0 != length ) )
    {
        m_LastError = USB_SERIAL_ERROR__DEVICE_NOT_OPENED;

        if ( m_PollFd.fd >= 0 )
        {
            m_LastError = USB_SERIAL_ERROR__SUCCESS;

            // Write all requested bytes
            unsigned int bytesToWrite = length;
            int bytesWritten;
            while ( bytesToWrite > 0 )
            {
                bytesWritten = ::write( m_PollFd.fd, pBuffer + length - bytesToWrite, bytesToWrite );
                if ( bytesWritten <= 0 )
                {
                    m_LastError = USB_SERIAL_ERROR__DEVICE_IO_OPERATION_ERROR;
                    break;
                }
                bytesToWrite -= bytesWritten;
            }
        }
    }

    return ( USB_SERIAL_ERROR__SUCCESS == m_LastError );
}

UsbSerial::ErrorCode UsbSerial::GetLastError() const
{
    return m_LastError;
}

const char* UsbSerial::GetLastErrorString() const
{
    switch ( m_LastError )
    {
    case USB_SERIAL_ERROR__SUCCESS:
        return "success";

    case USB_SERIAL_ERROR__DEVICE_ALREADY_OPENED:
        return "USB device already opened";

    case USB_SERIAL_ERROR__PARAMETER_ERROR:
        return "user parameter error";

    case USB_SERIAL_ERROR__DEVICE_NOT_FOUND:
        return "USB device not found";

    case USB_SERIAL_ERROR__DEVICE_OPEN_ERROR:
        return "can't open USB device";

    case USB_SERIAL_ERROR__DEVICE_NOT_OPENED:
        return "USB device is not opened";

    case USB_SERIAL_ERROR__DEVICE_IO_OPERATION_ERROR:
        return "USB device I/O operation error";

    case USB_SERIAL_ERROR__DEVICE_IO_OPERATION_TIMEOUT:
        return "USB device I/O operation timeout";

    default:
        return ""; // Shall never return
    }
}
