#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include "reader_interface.h"

static void debug_log_func( const char* subsystem_name_ptr, const char* debug_level_name_ptr, char* log_text_ptr )
{
    printf( "[%s%s%s\n   %s\n", debug_level_name_ptr, "]: subsystem: ", subsystem_name_ptr, log_text_ptr );
    fflush( stdout );
}

static void read_card( unsigned long serial_number )
{
    unsigned long card_number;
    Read_card_response_t read_card_response;
    if ( 0 == reader_read_card( serial_number, &card_number, &read_card_response ) )
    {
        printf( "%s%lu:\n", "Received response from reader with S/N ", serial_number );
        unsigned char* data_ptr = NULL;
        unsigned int data_length;
        switch ( read_card_response.card_type )
        {
        case CARD_TYPE__MIFARE:
            printf( "%s%u%s%u%s%u%s", "  MIFARE: ATQA ",
                    ( read_card_response.card.mifare.atqa.high << 8 ) + read_card_response.card.mifare.atqa.low,
                    ", SAK ", read_card_response.card.mifare.sak, ", UID length ", read_card_response.card.mifare.uid_length, ", UID (hex):" );
            data_ptr = &( read_card_response.card.mifare.uid[ 0 ] );
            data_length = read_card_response.card.mifare.uid_length;
            break;

        case CARD_TYPE__EM_MARINE:
        case CARD_TYPE__HID:
            printf( "%s%s", ( CARD_TYPE__HID == read_card_response.card_type ) ? "  HID:" : "  EM_MARINE:", " DATA (hex): " );
            data_ptr = &( read_card_response.card.em_marine_hid.data[ 0 ] );
            data_length = read_card_response.card.em_marine_hid.length;
            break;

        default:
            /* Will never be called */
            break;
        }

        if ( NULL != data_ptr )
        {
            unsigned int i;
            for ( i = 0; i < data_length; i++ )
            {
                printf( " %02X", *( data_ptr + i ) );
            }
            printf( "%s%08lX", ", card number: ", card_number );
            printf( "\n\n" );
        }
    }
}

int main( int argc, char* argv[] )
{
    if ( ( 3 != argc ) && ( 4 != argc ) )
    {
        printf( "%s%s%s", "\nUsage: ", argv[ 0 ], " <iterations> <debug_level> [weigand_26]\n" );
        printf( "%s", "  where:\n" );
        printf( "%s", "    iterations is the number of 1 second intervals before exit\n" );
        printf( "%s", "    debug_level value is between 0 and 3 (0 - minimum, 3 - maximum)\n" );
        printf( "%s", "    set weigand_26 to 1 to enable weigand 26\n\n" );
        exit( 0 );
    }

    unsigned int iterations = atoi( argv[ 1 ] );
    unsigned int debug_level = atoi( argv[ 2 ] );

    reader_set_wiegand_26( FALSE );
    if ( ( 4 == argc ) && ( argv[ 3 ] ) )
    {
        reader_set_wiegand_26( TRUE );
    }

    /* Set debug log function and debug levels: 0 - minimum, 3 - maximum */
    reader_set_log_func( debug_log_func );
    reader_debug_level( debug_level, debug_level, debug_level );

    /* Initialize reader interface */
    if ( 0 == reader_interface_init() )
    {
        unsigned long attachedReader[ 8 ];

        // Wait until fabric is initialized; could be skipped in real application as such condition is gracefully handled within the fabric
        usleep( 500000 );

        while ( iterations-- > 0 )
        {
            // Get attached reader list
            unsigned int attachedReaderIndex = reader_get_attached( &( attachedReader[ 0 ] ), sizeof ( attachedReader ) / sizeof ( attachedReader[ 0 ] ) );

            if ( attachedReaderIndex > 0 )
            {
                /* Go through attached readers */
                while ( attachedReaderIndex-- > 0 )
                {
                    read_card( attachedReader[ attachedReaderIndex ] );
                }
            }
            else
            {
                printf( "%s", "No attached readers found\n\n" );
            }

            usleep( 1000000 );
        }

        reader_interface_term();
    }
    else
    {
        printf( "%s", "Can't initialize reader interface\n" );
    }

    return 0;
}
